@@ -2,19 +2,29 @@ extern crate downcast;
 extern crate once_cell;
 extern crate variadic_generics;
 
+use async_trait::async_trait;
 use downcast::{AnySync, TypeMismatch};
 use once_cell::sync::OnceCell;
 use variadic_generics::va_expand;
-use std::error::{self, Error as StdError};
-use std::result::Result as StdResult;
-use std::sync::Arc;
 use std::collections::HashMap;
+use std::error::{self, Error as StdError};
 use std::fmt;
+use std::future::Future;
+use std::pin::Pin;
+use std::sync::Arc;
+
+/* Current TODOs:
+ *
+ * 1. write tests
+ * 2. figure out which parts of the library should be hidden
+ * 3. detect and prevent hangups caused by mutual dependencies
+ * 4. do we want TransientInstancers besides SingletonInstancers?
+ */
 
 // errors --------------------------------------------------
 
-pub type Error = Box<dyn StdError + 'static>;
-pub type Result<T> = StdResult<T, Error>;
+pub extern crate anyhow;
+pub use anyhow::{Result, Error};
 
 #[derive(Debug)]
 pub struct InstancerNotFoundError {
@@ -84,33 +94,44 @@ impl error::Error for InstanceTypeError {
 }
 
 
-// Resolve & ResolveStart --------------------------------------------------
+// Resolve, ResolveStart --------------------------------------------------
 
-pub trait Resolve: Sized + 'static {
-    type Deps;
+#[async_trait]
+pub trait Resolve: Send + Sized + 'static {
+    type Deps: Send;
     fn resolve(deps: Self::Deps) -> Result<Self>;
+    async fn resolve_async(deps: Self::Deps) -> Result<Self>;
 }
 
 /// Careful when using this trait, or you'll be in for a world of stack
 /// overflows.
-pub trait ResolveStart<R> {
+#[async_trait]
+pub trait ResolveStart<R>: Sync {
     fn resolve_start(&self) -> Result<R>;
+    async fn resolve_start_async(&self) -> Result<R>;
 }
 
-impl<X> ResolveStart<()> for X {
+#[async_trait]
+impl<X: Sync> ResolveStart<()> for X {
     fn resolve_start(&self) -> Result<()> { Ok(()) }
+    async fn resolve_start_async(&self) -> Result<()> { Ok(()) }
 }
 
+#[async_trait]
 impl<R, X> ResolveStart<R> for X
     where R: Resolve, X: ResolveStart<R::Deps>
 {
     fn resolve_start(&self) -> Result<R> {
         R::resolve(<X as ResolveStart<R::Deps>>::resolve_start(self)?)
     }
+    async fn resolve_start_async(&self) -> Result<R> {
+        R::resolve_async(<X as ResolveStart<R::Deps>>::resolve_start_async(self).await?).await
+    }
 }
 
 // tuples
 va_expand!{ ($va_len:tt) ($($va_idents:ident),+) ($($va_indices:tt),+)
+    #[async_trait]
     impl<$($va_idents,)+ X> ResolveStart<($($va_idents,)+)> for X
     where 
         $($va_idents: Resolve,)+
@@ -121,6 +142,11 @@ va_expand!{ ($va_len:tt) ($($va_idents:ident),+) ($($va_indices:tt),+)
                 $va_idents::resolve(<X as ResolveStart<$va_idents::Deps>>::resolve_start(self)?)?,
             )+))
         }
+        async fn resolve_start_async(&self) -> Result<($($va_idents,)+)> { 
+            Ok(($(
+                $va_idents::resolve_async(<X as ResolveStart<$va_idents::Deps>>::resolve_start_async(self).await?).await?,
+            )+))
+        }
     }
 }
 
@@ -144,26 +170,30 @@ pub struct InstantiationRequest {
 }
 
 impl InstantiationRequest {
-    pub fn increment_shadow(&mut self, service_name: &str){
+    fn increment_shadow(&mut self, service_name: &str){
         let level = self.shadow_levels.entry(service_name.to_owned())
             .or_insert(0);
         *level += 1;
     }
     
     /// returns true if successfully decremented
-    pub fn decrement_shadow(&mut self, service_name: &str) -> bool {
+    fn decrement_shadow(&mut self, service_name: &str) -> bool {
         self.shadow_levels.get_mut(service_name)
             .map(|level| level.saturating_sub(1))
             .unwrap_or(1) != 0
     }
 }
 
+#[async_trait]
 pub trait Middleware: Send + Sync + 'static {
     fn instantiate(&self, req: InstantiationRequest) -> Result<InstanceRef>;
+    async fn instantiate_async(&self, req: InstantiationRequest) -> Result<InstanceRef>;
 }
 
+#[async_trait]
 impl ResolveStart<Arc<dyn Middleware>> for Arc<dyn Middleware> {
     fn resolve_start(&self) -> Result<Arc<dyn Middleware>> { Ok(self.clone()) }
+    async fn resolve_start_async(&self) -> Result<Arc<dyn Middleware>> { Ok(self.clone()) }
 }
 
 // InstanceRef & TypedInstanceRef --------------------------------------------------
@@ -173,10 +203,12 @@ pub type InstanceRef = Arc<dyn AnySync>;
 #[allow(type_alias_bounds)]
 pub type TypedInstanceRef<S: ?Sized> = Arc<Box<S>>;
 
+#[async_trait]
 impl<S> Resolve for TypedInstanceRef<S>
     where S: Service + ?Sized
 {
     type Deps = Arc<dyn Middleware>;
+
     fn resolve(top: Self::Deps) -> Result<Self> {
         let req = InstantiationRequest{
             top: top.clone(),
@@ -187,31 +219,47 @@ impl<S> Resolve for TypedInstanceRef<S>
             .downcast_arc::<Box<S>>()
             .map_err(|err| InstanceTypeError::new(S::service_name(), err.type_mismatch()).into())
     }
+
+    async fn resolve_async(top: Self::Deps) -> Result<Self> {
+        let req = InstantiationRequest{
+            top: top.clone(),
+            service_name: S::service_name(),
+            shadow_levels: Some((S::service_name(), 1)).into_iter().collect(),
+        };
+        top.instantiate_async(req).await?
+            .downcast_arc::<Box<S>>()
+            .map_err(|err| InstanceTypeError::new(S::service_name(), err.type_mismatch()).into())
+    }
 }
 
 // Middleware: ContainerRoot --------------------------------------------------
 
-pub struct ContainerRoot;
+struct ContainerRoot;
 
+#[async_trait]
 impl Middleware for ContainerRoot {
     fn instantiate(&self, req: InstantiationRequest) -> Result<InstanceRef> {
         Err(InstancerNotFoundError::new(req.service_name).into())
     }
+    async fn instantiate_async(&self, req: InstantiationRequest) -> Result<InstanceRef> {
+        self.instantiate(req)
+    }
 }
 
-// Middleware: SingletonInstance, TransientInstancer  --------------------------------------------------
+// Middleware: InstancerShadow --------------------------------------------------
 
-pub struct InstancerShadow {
+struct InstancerShadow {
     prev: Arc<dyn Middleware>,
     shadowed_service_name: String
 }
 
 impl InstancerShadow {
-    pub fn new(prev: Arc<dyn Middleware>, shadowed_service_name: String) -> Self {
+    fn new(prev: Arc<dyn Middleware>, shadowed_service_name: String) -> Self {
         Self{ prev, shadowed_service_name }
     }
 }
 
+#[async_trait]
 impl Middleware for InstancerShadow {
     fn instantiate(&self, mut req: InstantiationRequest) -> Result<InstanceRef> {
         if self.shadowed_service_name == req.service_name {
@@ -219,14 +267,25 @@ impl Middleware for InstancerShadow {
         }
         self.prev.instantiate(req)
     }
+    async fn instantiate_async(&self, mut req: InstantiationRequest) -> Result<InstanceRef> {
+        if self.shadowed_service_name == req.service_name {
+            req.increment_shadow(&self.shadowed_service_name)
+        }
+        self.prev.instantiate_async(req).await
+    }
 }
 
-#[allow(type_alias_bounds)]
-pub type CreationFn<T: ?Sized> = Arc<dyn (Fn(&Arc<dyn Middleware>) -> Result<Box<T>>) + Send + Sync>;
+// Middleware: (Async)SingletonInstancer  --------------------------------------------------
 
-pub struct SingletonInstancer<T: ?Sized> {
+#[allow(type_alias_bounds)]
+type CreationFn<T: ?Sized> = Arc<dyn (Fn(&Arc<dyn Middleware>) -> Result<Box<T>>) + Send + Sync>;
+#[allow(type_alias_bounds)]
+type AsyncCreationFn<T: ?Sized> = Arc<dyn (Fn(&'_ Arc<dyn Middleware>) -> Pin<Box<dyn Future<Output = Result<Box<T>>> + Send + '_>>) + Send + Sync>;
+
+struct SingletonInstancer<T: ?Sized> {
     prev: Arc<dyn Middleware>,
     creation_fn: CreationFn<T>,
+    #[allow(clippy::redundant_allocation)]
     instance: OnceCell<Arc<Box<T>>>,
     service_name: String,
 }
@@ -234,12 +293,13 @@ pub struct SingletonInstancer<T: ?Sized> {
 impl<T> SingletonInstancer<T>
     where T: Service + ?Sized
 {
-    pub fn new(prev: Arc<dyn Middleware>, creation_fn: CreationFn<T>) -> Self {
+    fn new(prev: Arc<dyn Middleware>, creation_fn: CreationFn<T>) -> Self {
         let service_name = T::service_name();
         Self{ prev, creation_fn, instance: OnceCell::new(), service_name } 
     }
 }
 
+#[async_trait]
 impl<T> Middleware for SingletonInstancer<T>
     where T: Service + ?Sized
 {
@@ -260,27 +320,41 @@ impl<T> Middleware for SingletonInstancer<T>
             .map(|inst| inst.clone() as Arc<dyn AnySync>)
             .map_err(|err| InstanceCreationError::new(self.service_name.clone(), err).into())
     }
-}
-
-pub struct TransientInstancer<T: ?Sized> {
-    prev: Arc<dyn Middleware>,
-    creation_fn: CreationFn<T>,
-    service_name: String,
-}
-
-impl<T> TransientInstancer<T>
-    where T: Service + ?Sized
-{
-    pub fn new(prev: Arc<dyn Middleware>, creation_fn: CreationFn<T>) -> Self {
-        let service_name = T::service_name();
-        Self{ prev, creation_fn, service_name } 
+    
+    async fn instantiate_async(&self, req: InstantiationRequest) -> Result<InstanceRef> {
+        self.instantiate(req)
     }
 }
 
-impl<T> Middleware for TransientInstancer<T>
+struct AsyncSingletonInstancer<T: ?Sized> {
+    prev: Arc<dyn Middleware>,
+    creation_fn: AsyncCreationFn<T>,
+    #[allow(clippy::redundant_allocation)]
+    instance: futures::lock::Mutex<Option<Arc<Box<T>>>>,
+    service_name: String,
+}
+
+impl<T> AsyncSingletonInstancer<T>
     where T: Service + ?Sized
 {
-    fn instantiate(&self, mut req: InstantiationRequest) -> Result<InstanceRef> {
+    fn new(prev: Arc<dyn Middleware>, creation_fn: AsyncCreationFn<T>) -> Self {
+        let service_name = T::service_name();
+        Self{ prev, creation_fn, instance: futures::lock::Mutex::new(None), service_name } 
+    }
+}
+
+#[async_trait]
+impl<T> Middleware for AsyncSingletonInstancer<T>
+    where T: Service + ?Sized
+{
+    fn instantiate(&self, req: InstantiationRequest) -> Result<InstanceRef> {
+        Err(InstanceCreationError::new(
+                req.service_name, 
+                anyhow::format_err!("cannot call async creation function from non-async code, use async functions instead")
+        ).into())
+    }
+    
+    async fn instantiate_async(&self, mut req: InstantiationRequest) -> Result<InstanceRef> {
         // if different service or shadowed, pass request (with one less shadow level) up the chain
         if req.service_name != self.service_name
         || req.decrement_shadow(&self.service_name)
@@ -292,10 +366,15 @@ impl<T> Middleware for TransientInstancer<T>
         req.increment_shadow(&self.service_name);
         let shadowed_top: Arc<dyn Middleware> = Arc::new(InstancerShadow::new(req.top, self.service_name.clone()));
         
-        // create instance
-        (self.creation_fn)(&shadowed_top)
-            .map(|inst| Arc::new(inst) as InstanceRef)
-            .map_err(|err| InstanceCreationError::new(self.service_name.clone(), err).into())
+        // recall or create instance
+        let mut guard = self.instance.lock().await;
+        if guard.is_none() {
+            let inst = (self.creation_fn)(&shadowed_top).await
+                .map(Arc::new)
+                .map_err(|err| InstanceCreationError::new(self.service_name.clone(), err))?;
+            *guard = Some(inst);
+        }
+        Ok(guard.as_ref().cloned().unwrap())
     }
 }
 
@@ -306,11 +385,15 @@ pub struct Container {
     top: Arc<dyn Middleware>,
 }
 
+#[async_trait]
 impl Resolve for Container {
     type Deps = Arc<dyn Middleware>;
     fn resolve(top: Self::Deps) -> Result<Self> {
         Ok(Container{ top })
     }
+    async fn resolve_async(top: Self::Deps) -> Result<Self> {
+        Ok(Container{ top })
+    }
 }
 
 impl Default for Container {
@@ -322,34 +405,58 @@ impl Container {
         Self{ top }
     }
 
-    pub fn with_singleton<S, Args>(&self, f: impl Fn(Args) -> Result<Box<S>> + Send + Sync + 'static) -> Self
-        where S: Service + ?Sized, Arc<dyn Middleware>: ResolveStart<Args>
+    pub fn with_singleton<S, Args, F>(&self, creation_fn: F) -> Self
+    where 
+        S: Service + ?Sized,
+        Arc<dyn Middleware>: ResolveStart<Args>,
+        F: Fn(Args) -> Result<Box<S>> + Send + Sync + 'static
     {
-    	let creation_fn = move |mw: &Arc<dyn Middleware>| f(mw.resolve_start()?);
-        Self::new(Arc::new(SingletonInstancer::new(self.top.clone(), Arc::new(creation_fn))))
+    	let creation_fn: CreationFn<S> = Arc::new(move |mw: &Arc<dyn Middleware>| {
+            creation_fn(mw.resolve_start()?)
+        });
+        Self::new(Arc::new(SingletonInstancer::new(self.top.clone(), creation_fn)))
     }
 
-    pub fn with_singleton_ok<S, Args>(&self, f: impl Fn(Args) -> Box<S> + Send + Sync + 'static) -> Self
-        where S: Service + ?Sized, Arc<dyn Middleware>: ResolveStart<Args>
+    pub fn with_singleton_ok<S, Args, F>(&self, creation_fn: F) -> Self
+    where
+        S: Service + ?Sized,
+        Arc<dyn Middleware>: ResolveStart<Args>,
+        F: Fn(Args) -> Box<S> + Send + Sync + 'static
     {
-    	let creation_fn = move |mw: &Arc<dyn Middleware>| Ok(f(mw.resolve_start()?));
-        Self::new(Arc::new(SingletonInstancer::new(self.top.clone(), Arc::new(creation_fn))))
+    	let creation_fn: CreationFn<S> = Arc::new(move |mw: &Arc<dyn Middleware>| {
+            Ok(creation_fn(mw.resolve_start()?))
+        });
+        Self::new(Arc::new(SingletonInstancer::new(self.top.clone(), creation_fn)))
     }
 
-    pub fn with_transient<S, Args>(&self, f: impl Fn(Args) -> Result<Box<S>> + Send + Sync + 'static) -> Self
-        where S: Service + ?Sized, Arc<dyn Middleware>: ResolveStart<Args>
+    pub fn with_singleton_async<S, Args, Fut, F>(&self, creation_fn: F) -> Self
+    where 
+        S: Service + ?Sized,
+        Arc<dyn Middleware>: ResolveStart<Args>,
+        Args: Send,
+        Fut: Future<Output = Result<Box<S>>> + Send,
+        F: Fn(Args) -> Fut + Send + Sync + Copy + 'static
     {
-    	let creation_fn = move |mw: &Arc<dyn Middleware>| f(mw.resolve_start()?);
-        Self::new(Arc::new(TransientInstancer::new(self.top.clone(), Arc::new(creation_fn))))
+    	let creation_fn: AsyncCreationFn<S> = Arc::new(move |mw| {
+            Box::pin(async move { creation_fn(mw.resolve_start_async().await?).await })
+        });
+        Self::new(Arc::new(AsyncSingletonInstancer::<S>::new(self.top.clone(), creation_fn)))
     }
 
-    pub fn with_transient_ok<S, Args>(&self, f: impl Fn(Args) -> Box<S> + Send + Sync + 'static) -> Self
-        where S: Service + ?Sized, Arc<dyn Middleware>: ResolveStart<Args>
+    pub fn with_singleton_async_ok<S, Args, Fut, F>(&self, creation_fn: F) -> Self
+    where 
+        S: Service + ?Sized,
+        Arc<dyn Middleware>: ResolveStart<Args>,
+        Args: Send,
+        Fut: Future<Output = Box<S>> + Send,
+        F: Fn(Args) -> Fut + Send + Sync + Copy + 'static
     {
-    	let creation_fn = move |mw: &Arc<dyn Middleware>| Ok(f(mw.resolve_start()?));
-        Self::new(Arc::new(TransientInstancer::new(self.top.clone(), Arc::new(creation_fn))))
+    	let creation_fn: AsyncCreationFn<S> = Arc::new(move |mw| {
+            Box::pin(async move { Ok(creation_fn(mw.resolve_start_async().await?).await) })
+        });
+        Self::new(Arc::new(AsyncSingletonInstancer::<S>::new(self.top.clone(), creation_fn)))
     }
-    
+
     pub fn resolve<X>(&self) -> Result<X>
         where Arc<dyn Middleware>: ResolveStart<X>
     {
